/*
 *  typespeed - measures your typing speed
 *  Copyright (C) 1999-2003   Jani Ollikainen  <bestis@iki.fi>
 *                          & Jaakko Manelius  <jman@iki.fi>
 *  Copyright (C) 2006   Tobias Stoeckmann  <tobias@bugol.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * convert.c - convert your highscore for 0.5.0 and later
 */

#include <sys/types.h>
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

struct list {
	int fidelity;
	int pong;
	float cps;
	float tcps;
	float typoratio;
	char faim[21];
};

void usage(const char *);

/*
 * convert will open a highscore file if it has the correct size,
 * else the old way of reading stuff can fail. Also we can be
 * rather sure that it must be a highscore file ...
 *
 * The new highscore will be written into same file without
 * closing the stream, this way we can be sure that the opened
 * file won't be modified during our operations.
 *
 * Returns 0 on success or 1 on failure.
 */
int
main(int argc, char **argv)
{
	FILE *highf;
	int i;
	struct list high[10];
	struct stat status;

	memset(&high, 0, sizeof(high));
	if (argc != 2)
		usage(argv[0]);

	if (stat(argv[1], &status)) {
		fprintf(stderr, "Cannot retrieve status of %s\n", argv[1]);
		exit(1);
	}
	if (status.st_size != 410) {
		fprintf(stderr, "%s: no valid highscore file!\n", argv[1]);
		exit(1);
	}
	if ((highf = fopen(argv[1], "r+")) == NULL) {
		fprintf(stderr, "Cannot open %s for modifying!\n", argv[1]);
		exit(1);
	}

	for (i = 0; i < 10; i++) {
		fread(&high[i].pong, sizeof(int), 1, highf);
		fread(&high[i].cps, sizeof(float), 1, highf);
		fread(&high[i].fidelity, sizeof(int), 1, highf);
		fread(&high[i].tcps, sizeof(float), 1, highf);
		fread(&high[i].typoratio, sizeof(float), 1, highf);
		fread(&high[i].faim, sizeof(char), 21, highf);
	}
	rewind(highf);
	if (fwrite(&high, sizeof(struct list), 10, highf) != 10) {
		fprintf(stderr, "Error during write!");
		exit(1);
	}

	fclose(highf);
	return 0;
}

/* Prints correct usage and exits with return value 1. */
void
usage(const char *program)
{
	fprintf(stderr, "usage: %s highscorefile\n", program);
	exit(1);
}

