/*
 *  typespeed - measures your typing speed
 *  Copyright (C) 1999-2003   Jani Ollikainen  <bestis@iki.fi>
 *                          & Jaakko Manelius  <jman@iki.fi>
 *  Copyright (C) 2006   Tobias Stoeckmann  <tobias@bugol.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * misc.c - misc functions
 */

#include <sys/time.h>
#include <sys/types.h>

#include <curses.h>
#include <ctype.h>
#include <err.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "typespeed.h"

void	 endcursestuff(void);
void	 intcursestuff(void);
void	 liima_mvgetnstr(int, int, char *, int);
int	 level(int);
int	 r(int);
clock_t	 timenow(void);
int	 typorankkaus(float);
void	 xcolor_set(short);
void	*xmalloc(size_t);

/* Closes all relevant curses objects. */
void
endcursestuff(void)
{
	clear();
	refresh();
	endwin();
}

/* Initialises curses */
void
initcursestuff(void)
{
	initscr();
	noecho();

	start_color();
	flushinp();

	init_pair(1, COLOR_GREEN, COLOR_BLACK);
	init_pair(2, COLOR_WHITE, COLOR_BLACK);
	init_pair(3, COLOR_RED, COLOR_BLACK);
	init_pair(4, COLOR_MAGENTA, COLOR_BLACK);
	init_pair(5, COLOR_CYAN, COLOR_BLACK);
	init_pair(6, COLOR_BLUE, COLOR_BLACK);
	init_pair(7, COLOR_YELLOW, COLOR_BLACK);
}

/* Returns (0 - 10) and sets color according to "pointsit". */
int
level(int pointsit)
{
	int leveli;

	if (pointsit < 0)
		leveli = 0;
	else if ((leveli = pointsit / 100 + 1) > 10)
		leveli = 10;

	if (pointsit < 400)
		xcolor_set(6);
	else if (pointsit < 600)
		xcolor_set(1);
	else if (pointsit < 700)
		xcolor_set(7);
	else
		xcolor_set(3);

	return leveli;
}

/*
 * yet another purkka by jaakko ..
 * (no mvgetnstr support in jaakko's ncurses)
 */
void
liima_mvgetnstr(int y, int x, char *buf, int maxlen)
{
	int curlen = 0;
	int merkki;

	memset(buf, 0, maxlen + 1);
	for (;;) {
		refresh();
		move(y, x + curlen);
		addstr("      ");

		move(y, x);
		if (curlen)
			addstr(buf);
		refresh();

		while ((merkki = getch()) == ERR)
			;
		if (merkki == '\n' || merkki == KEY_ENTER)
			break;
		switch (merkki) {
		case 27:	/* ESC */
			flushinp();
			break;
		case 8:		/* BS */
		case 127:	/* DEL */
		case KEY_BACKSPACE:
		case 4:		/* EOT */
			if (curlen) {
				echochar(merkki);
				curlen--;
				buf[curlen] = '\0';
			}
			break;
		default:
			if (curlen != maxlen && !iscntrl(merkki)) {
				buf[curlen] = merkki;
				curlen++;
			}
			break;
		}
	}
}

/* Returns a random number between 0 and "range". */
int
r(int range)
{
	return (int)(random() % range);
}

/* Returns the current time in hundrets of seconds. */
clock_t
timenow(void)
{
	struct timeval tval;
	gettimeofday(&tval, NULL);
	return ((clock_t)((tval.tv_sec * 100) + (tval.tv_usec / 10000)));
}

/* Returns (0 - 11) and sets color appropiate to "typorate". */
int
typorankkaus(float typorate)
{
	int typorankki;
	/*
	 * if you get a negative typorate here, you must
	 * be as good as my friend xmunkki
	 *
	 * and btw. thanks to xmunkki for his
	 * "OverHumanly debugging without compiler"
	 */
	if (typorate < 0)
		typorankki = 0;
	else if (typorate == 0)
		typorankki = 1;
	else if (typorate < 2)
		typorankki = 2;
	else if (typorate < 4)
		typorankki = 3;
	else if (typorate < 6)
		typorankki = 4;
	else if (typorate < 8)
		typorankki = 5;
	else if (typorate < 11)
		typorankki = 6;
	else if (typorate < 15)
		typorankki = 7;
	else if (typorate < 20)
		typorankki = 8;
	else if (typorate < 30)
		typorankki = 9;
	else if (typorate < 50)
		typorankki = 10;
	else
		typorankki = 11;

	if (typorate < 6)
		xcolor_set(6);
	else if (typorate < 11)
		xcolor_set(1);
	else if (typorate < 20)
		xcolor_set(7);
	else
		xcolor_set(3);

	return typorankki;
}

/* Calls color_set if opt.usecolors is set. */
void
xcolor_set(short color_pair_number)
{
	if (opt.usecolors == 1)
		color_set(color_pair_number, NULL);
}

/* Wrapper for malloc */
void *
xmalloc(size_t size)
{
	void *pointer;
	if ((pointer = malloc(size)) == NULL)
		err(1, "malloc()");
	return pointer;
}

